Spell = {}
SPELL_WATER_LEVEL = 1500
SPELL_OPACITY = 0.65
SPELL_DURATION_HIDE_THRESHOLD = 20
DOT = "dot"
DEBUFF = "debuff"
BUFF = "buff"
PROTO_LINKABLE = "linkable"
LINK_TIMELEFT_CEILING = 10000

--[[
Some fields that might not be obvious
applied_by = Buff and Debuff fields have this property. It's for a buff/debuff that's applied by another spell. 
This is used to get things like the cast time of the spell so the icons change color at the right times.

applied_by_id = the id of the applied spell for translation

prototype = "linkable" - this links a spell slot to another slot. If two spells are mutually exclusive you would 
want them to take up the same slot in the UI. prototype = linkable gives you this.

link_to = spell_id - paired with prototype, links the spell.
]]--
function Spell:new(name, id, props, texture, timefont, linkedSpell)
	local object = { props = props, id = id, castTime = 0, name = name, texture = texture, timefont = timefont,
					timeleft = 0, stacks = 0, shown = false, shader = nil, tex_verify = false,
					displayTimeleft = 0, disableStacks = props.disable_stacks or false, linkedSpell = linkedSpell,
					totalDuration = 0, linkTimeleft = LINK_TIMELEFT_CEILING, action_button_id = nil,
					disable_glow = props.disable_glow or false, firstApplication = true }
	object.shader = SpellShader:new(object);
	-- reset to yellow
	timefont:SetTextColor(1, 0.82, 0, 1);
	texture:SetDesaturated(nil);
	texture:SetAlpha(SPELL_OPACITY);
	setmetatable (object, { __index = Spell } )
	return object
end

function Spell:refresh(duration, timeleft, stacks) 
	if self.linkedSpell then 
		if timeleft < self.linkedSpell.linkTimeleft then
			self.linkedSpell.linkTimeleft = timeleft;
			self.linkedSpell.linkActive = self;
			
			self:updateDuration( duration, timeleft );
			self:updateStacks( stacks );
			self:shade();
			
			self.texture:Show();
			self.timefont:Show();
		else 
			self.texture:Hide();
			self.timefont:Hide();
		end
	else 
		--standard spell, not stackable
		self:updateDuration( duration, timeleft );
		self:updateStacks( stacks );
		self:shade();
	end
end

-- timeleft in seconds
function Spell:updateDuration(duration, timeleft)
	if timeleft < -1 then 
		-- fix for auras without a time
		timeleft = SPELL_DURATION_HIDE_THRESHOLD 
	end
	self.displayTimeleft = round(timeleft);
	timeleft = timeleft * 1000;
	
	--check for re-application
	if timeleft > self.timeleft then
		if self.props.spell_type == DOT then
			if self.id ~= nil and self.id ~= 0 then
				self:reapplySpell(self.id, duration);
			else
				self:reapplySpell(self.name, duration);
			end
		elseif self.props.spell_type == DEBUFF or self.props.spell_type == BUFF  then
			self:reapplySpell(self.props.applied_by, duration);
		end
	end
	self.timeleft = timeleft;
	
	if self.displayTimeleft < 0 then
		self.displayTimeleft = 0;
	elseif self.displayTimeleft < 5 then
		self.displayTimeleft = string.format("%1.1f", self.displayTimeleft);
	elseif self.displayTimeleft >= SPELL_DURATION_HIDE_THRESHOLD then
		self.displayTimeleft = "";
	else
		self.displayTimeleft = string.format("%1d", self.displayTimeleft);
	end
	
	self.timefont:SetText(self.displayTimeleft);	
end

function Spell:reapplySpell(spellId, duration)
	local castTime = select(4, GetSpellInfo(spellId));

	self.props.castTime = castTime;
	if self.firstApplication == true then
	-- only change duration if it's the original duration, not extended duration
	-- as when an Immolate is reapplied 3 seconds before it expires (it's duration goes from 15s to 19s).
		self.totalDuration = duration * 1000;
		self.firstApplication = false;
	end
end

function round(val)
	return floor((val + 0.05)*10)/10
end

--override to use the timefont instead
function Spell:updateStacks(stacks)
	if (stacks > 0 or self.stacks > 0) and not self.disableStacks then
		self.stacks = stacks;
		--if self:ignoreSpellProcessing() then
		--	return;
		--end
		self.timefont:SetText(self.stacks);	
		self.timefont:SetTextColor(1,0,0,1);
	end
end

function Spell:Show()
	self.shown = true;
	self.texture:Show();
	self.timefont:Show();
end

function Spell:Hide()
	if self.linkedSpell and self == self.linkedSpell.linkActive then
		self.linkedSpell.linkTimeleft = LINK_TIMELEFT_CEILING;
	end
	self.shown = false;
	self.texture:Hide();
	self.timefont:Hide();
end

function Spell:IsShown()
	return self.shown;
end

function Spell:shade()	
	self.shader:doShade();		
end

function Spell:removeShading()
	self.shader:glow(false);
end	

function Spell:unbindActionButton() 
	self.action_button_id = nil;
	self.action_button_spell_id = nil;
	self.action_button_slot_id = nil;
end

-- where the id is 1 through 120, corresponding to GetActionInfo(id)
function Spell:bindActionButton(action_id, action_spell_id) 
	if action_id <= 12 then
		self.action_button_id = "ActionButton"..action_id;
		self.action_button_spell_id = action_spell_id;
	elseif action_id <= 24 then
		-- unsupported right now, have to add more state to track the 2nd level of the primary bar
		-- self.action_bar_id = "ActionButton"..(action_id-12);
	elseif action_id <= 36 then
		self.action_button_id = "MultiBarRightButton"..(action_id-24);
		self.action_button_spell_id = action_spell_id;
	elseif action_id <= 48 then
		self.action_button_id = "MultiBarLeftButton"..(action_id-36);
		self.action_button_spell_id = action_spell_id;
	elseif action_id <= 60 then
		self.action_button_id = "MultiBarBottomRightButton"..(action_id-48);
		self.action_button_spell_id = action_spell_id;
	elseif action_id <= 72 then
		self.action_button_id = "MultiBarBottomLeftButton"..(action_id-60);
		self.action_button_spell_id = action_spell_id;
	end
	self.action_button_slot_id = action_id;
end

function Spell:getTargetDistance() 
	if not CheckInteractDistance("target", 1) then -- target farther than 28 yards
		return 34; -- target farther than 28 yards, so take average of 40-28/2 + 28 (guesstimate)
	elseif not CheckInteractDistance("target",2) then -- target farther than 11 yards, but closer than 28
		return 20; -- target is farther than 11 yards
	else 
		return 6; -- target closer than 11 yards
	end
end